import struct
import os
import cython
cimport cython
import numpy as np
cimport numpy as np
import sys

from cpython.bytearray cimport PyByteArray_AsString
from libc.stdio cimport FILE, fopen, fread, fclose, fwrite


DEF FORMAT_KEY = b"v3d_volume_pkbitdf_encod"
DEF HEADER_SIZE = 43
DEF COMPRESSION_ENLARGEMENT = 2
DEF LITTLE = b'L'
DEF REPEAT_MAX_LEN  = 255 - 222

cdef unsigned char[3] MAX_LEN = [79 - 31, 182 - 79, 222 - 182]
cdef double[3] MAX_EFF = [16. / 3., 16. / 4., 16. / 5.]
cdef char[3][2] ran = [[-3, 4], [-7, 8], [-15, 16]]
cdef unsigned char[3] shift_bits = [3, 4, 5]
cdef unsigned char[3] gap = [31, 79, 182]
cdef unsigned char[3] mask = [0b00000111, 0b0001111, 0b00011111]


cdef class PBD:
    """
    Supporting most biological image LOSSLESS compression with sound SNR, where blocks of the signal are close in intensity.
    It supports 16bit and 8bit compression, can compress up to 25% or even lower. Note that the 16bit used by other sources
    is incomplete and can only reach 50%. This package gives you a choice of a full blood one that can make 25%, but the
    full blood output might not be loaded by other programs currently.

    The compression/decompression are optimized and can even be faster than Vaa3D.

    modified from v3d_external/v3d_main/neuron_annotator/utility/ImageLoaderBasic.cpp

    by Zuohan Zhao, Southeast University

    2022/6/23
    """
    cdef:
        long long total_read_bytes, compression_pos, decompression_pos, read_step_size_bytes
        bytearray compression_buffer, decompression_buffer
        bint endian_switch, pbd16_full_blood
        bytes endian_sys

    def __init__(self, pbd16_full_blood=True, read_step_size_bytes = 1024 * 20000):
        """
        :param pbd16_full_blood: Turn off or on to allow the full blood saving of 16bit image loading. Note
         other programs may not be able to load it. Default is on. Default as turned on.
        :param read_step_size_bytes: Adjust the number of bytes for each time of buffer loading, default as 20000KB.
        """
        self.endian_sys = sys.byteorder[0].upper().encode('ascii')
        self.endian_switch = False
        self.decompression_pos = self.compression_pos = self.total_read_bytes = 0
        self.compression_buffer = self.decompression_buffer = bytearray()
        self.pbd16_full_blood = pbd16_full_blood
        self.read_step_size_bytes = read_step_size_bytes

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef long long decompress_pbd8(self, long long look_ahead):
        cdef:
            unsigned char* decomp = <unsigned char*>PyByteArray_AsString(self.decompression_buffer)
            unsigned char* comp = <unsigned char*>PyByteArray_AsString(self.compression_buffer)
            long long cp = self.compression_pos, dp = self.decompression_pos
            unsigned char count, shift, carry
            char delta
        while cp < look_ahead:
            count = comp[cp]
            cp += 1
            if count < 33:
                count += 1
                for shift in range(count):
                    decomp[dp + shift] = comp[cp + shift]
                cp += count
                dp += count
            elif count < 128:
                count -= 32
                shift = 0
                while count > 0:
                    if shift == 0:
                        carry = comp[cp]
                        cp += 1
                    delta = (carry & 0b00000011 << shift) >> shift
                    if delta == 3:
                        delta = -1
                    decomp[dp] = decomp[dp - 1] + delta
                    dp += 1
                    count -= 1
                    shift = shift + 2 & 7
            else:
                count -= 127
                for shift in range(count):
                    decomp[dp + shift] = comp[cp]
                dp += count
                cp += 1
        return dp

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef long long decompress_pbd16(self, long long look_ahead):
        cdef:
            unsigned char * decomp = <unsigned char *> PyByteArray_AsString(self.decompression_buffer)
            unsigned char * comp = <unsigned char *> PyByteArray_AsString(self.compression_buffer)
            long long cp = self.compression_pos, dp = self.decompression_pos
            unsigned char count, i
            char delta, shift
            unsigned short carry
            unsigned short* ptr

        while cp < look_ahead:
            count = comp[cp]
            cp += 1
            if count < 32:
                count = count + 1 << 1
                if self.endian_switch:
                    for i in range(0, count, 2):
                        decomp[dp + i] = comp[cp + i + 1]
                        decomp[dp + i + 1] = comp[cp + i]
                else:
                    for i in range(0, count, 2):
                        decomp[dp + i] = comp[cp + i]
                        decomp[dp + i + 1] = comp[cp + i + 1]
                cp += count
                dp += count
                continue
            elif count < 80:
                i = 0
            elif count < 183:
                i = 1
            elif count < 223:
                i = 2
            else:
                count = count - 222 << 1
                if self.endian_switch:
                    for i in range(0, count, 2):
                        decomp[dp + i] = comp[cp + 1]
                        decomp[dp + i + 1] = comp[cp]
                else:
                    for i in range(0, count, 2):
                        decomp[dp + i] = comp[cp]
                        decomp[dp + i + 1] = comp[cp + 1]
                dp += count
                cp += 2
                continue
            count -= gap[i]
            shift = 0
            ptr = <unsigned short *> &decomp[dp]
            while count > 0:
                shift -= shift_bits[i]
                if shift < 0:
                    carry = carry << 8 | comp[cp]
                    cp += 1
                    shift += 8
                delta = carry >> shift & mask[i]
                if delta > ran[i][1]:
                    delta = ran[i][1] - delta
                ptr[0] = ptr[-1] + delta
                ptr += 1
                dp += 2
                count -= 1
        return dp

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void update_compression_buffer8(self):
        cdef:
            long long look_ahead = self.compression_pos
            unsigned char lav, compressed_diff_entries
        while look_ahead < self.total_read_bytes:
            lav = self.compression_buffer[look_ahead]
            if lav < 33:
                if look_ahead + lav + 1 < self.total_read_bytes:
                    look_ahead += lav + 2
                else:
                    break
            elif lav < 128:
                compressed_diff_entries = (lav - 33) // 4 + 1
                if look_ahead + compressed_diff_entries < self.total_read_bytes:
                    look_ahead += compressed_diff_entries + 1
                else:
                    break
            else:
                if look_ahead + 1 < self.total_read_bytes:
                    look_ahead += 2
                else:
                    break
        self.decompression_pos = self.decompress_pbd8(look_ahead)
        self.compression_pos = look_ahead

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void update_compression_buffer16(self):
        cdef:
            long long look_ahead = self.compression_pos
            unsigned char lav, compressed_diff_bytes
        while look_ahead < self.total_read_bytes:
            lav = self.compression_buffer[look_ahead]
            if lav < 32:
                if look_ahead + (lav + 1) * 2 < self.total_read_bytes:
                    look_ahead += (lav + 1) * 2 + 1
                else:
                    break
            elif lav < 80:
                compressed_diff_bytes = ((lav - 31) * 3 - 1) // 8 + 1
                if look_ahead + compressed_diff_bytes < self.total_read_bytes:
                    look_ahead += compressed_diff_bytes + 1
                else:
                    break
            elif lav < 183:
                compressed_diff_bytes = ((lav - 79) * 4 - 1) // 8 + 1
                if look_ahead + compressed_diff_bytes < self.total_read_bytes:
                    look_ahead += compressed_diff_bytes + 1
                else:
                    break
            elif lav < 223:
                compressed_diff_bytes = ((lav - 182) * 5 - 1) // 8 + 1
                if look_ahead + compressed_diff_bytes < self.total_read_bytes:
                    look_ahead += compressed_diff_bytes + 1
                else:
                    break
            else:
                if look_ahead + 2 < self.total_read_bytes:
                    look_ahead += 3
                else:
                    break
        self.decompression_pos = self.decompress_pbd16(look_ahead)
        self.compression_pos = look_ahead

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cpdef np.ndarray load(self, path: str | os.PathLike):
        """
        :param path: output image path of v3dpbd.
        :return: a 4D numpy array of either uint8 or uint16.
        """
        file_size = os.path.getsize(path)
        assert file_size >= HEADER_SIZE, "File size smaller than header size."
        cdef:
            short datatype
            long long current_read_bytes, channel_len, remaining_bytes
            const unsigned char[:] p = str(path).encode('utf-8')
            FILE* f = fopen(<const char*>&p[0], <char*>'rb')
        if f is NULL:
            raise Exception("Fail to open file for reading.")
        try:
            header = bytearray(HEADER_SIZE)
            fread(PyByteArray_AsString(header), HEADER_SIZE, 1, f)
            assert header.find(FORMAT_KEY) == 0, "Format key loading failed."
            header = header[len(FORMAT_KEY):]
            endian = '<' if header[:1] == LITTLE else '>'
            self.endian_switch = header[:1] != self.endian_sys
            datatype = struct.unpack(f'{endian}h', header[1:3])[0]
            assert datatype in [1, 2], "Datatype can only be 1 or 2."
            sz = struct.unpack(f'{endian}iiii', header[3:])
            channel_len = sz[0] * sz[1] * sz[2]
            remaining_bytes = file_size - HEADER_SIZE
            self.total_read_bytes = 0
            self.compression_buffer = bytearray(remaining_bytes)
            self.decompression_buffer = bytearray(channel_len * sz[3] * datatype)
            self.compression_pos = self.decompression_pos = 0
            while remaining_bytes > 0:
                current_read_bytes = min(remaining_bytes, self.read_step_size_bytes,
                                         (self.total_read_bytes // channel_len + 1) *
                                         channel_len - self.total_read_bytes)
                fread(PyByteArray_AsString(self.compression_buffer) + self.total_read_bytes, current_read_bytes, 1, f)
                self.total_read_bytes += current_read_bytes
                remaining_bytes -= current_read_bytes
                if datatype == 1:
                    self.update_compression_buffer8()
                elif datatype == 2:
                    self.update_compression_buffer16()
                else:
                    raise Exception("Invalid datatype")
            return np.frombuffer(self.decompression_buffer, f'{endian}u{datatype}').reshape(sz[::-1])
        finally:
            fclose(f)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef long long compress_pbd8(self):
        cdef:
            unsigned char cur_val, prior_val, retest
            unsigned char[96] dbuffer
            unsigned char * decomp = <unsigned char *> PyByteArray_AsString(self.decompression_buffer)
            unsigned char * comp = <unsigned char *> PyByteArray_AsString(self.compression_buffer)
            short delta
            long long active_literal_index = -1, cur_dp, cp = 0, dp = 0
            double re_efficiency, df_efficiency
        assert self.decompression_pos > 0, "The buffer to save is empty."
        while dp < self.decompression_pos:
            if cp >= self.compression_pos:
                raise Exception("compression running out of space, try enlarging the compression buffer.")
            retest = 1
            cur_val = decomp[dp]
            cur_dp = dp + 1
            while cur_dp < self.decompression_pos and retest < 128 and decomp[cur_dp] == cur_val:
                retest += 1
                cur_dp += 1
            re_efficiency = retest / 2.

            if re_efficiency < 4:
                df_efficiency = 0.
                cur_dp = dp
                if dp > 0:
                    prior_val = decomp[dp - 1]
                    for cur_dp in range(dp, dp + min(self.decompression_pos - dp, 95)):
                        delta = decomp[cur_dp] - prior_val
                        if delta > 2 or delta < -1:
                            break
                        prior_val = decomp[cur_dp]
                        if delta == -1:
                            delta = 3
                        dbuffer[cur_dp - dp] = delta
                    else:
                        cur_dp += 1
                    df_efficiency = (cur_dp - dp) / ((cur_dp - dp) / 4. + 2)
            if re_efficiency >= 4. or re_efficiency > df_efficiency and re_efficiency > 1.:
                comp[cp] = retest + 127
                cp += 1
                comp[cp] = cur_val
                cp += 1
                active_literal_index = -1
                dp += retest
            elif df_efficiency > 1.:
                comp[cp] = cur_dp - dp + 32
                cp += 1
                for delta in range(0, cur_dp - dp, 4):
                    comp[cp] = dbuffer[delta+3] << 6 | dbuffer[delta+2] << 4 | dbuffer[delta+1] << 2 | dbuffer[delta]
                    cp += 1
                active_literal_index = -1
                dp = cur_dp
            else:
                if active_literal_index < 0 or comp[active_literal_index] >= 32:
                    comp[cp] = 0
                    active_literal_index = cp
                    cp += 1
                else:
                    comp[active_literal_index] += 1
                comp[cp] = cur_val
                cp += 1
                dp += 1
        return cp

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef long long compress_pbd16(self):
        cdef:
            unsigned char * decomp = <unsigned char *> PyByteArray_AsString(self.decompression_buffer)
            unsigned char * comp = <unsigned char *> PyByteArray_AsString(self.compression_buffer)
            unsigned char retest, carry, i = 0
            char shift
            unsigned char* pb
            unsigned short* pcp2
            unsigned short cur_val, prior_val
            const unsigned short* decomp2 = <unsigned short*>&decomp[0]
            long long decomp_len = self.decompression_pos // 2, active_literal_index = -1, dp2 = 0, cp = 0, cur_dp2
            int delta
            double re_efficiency
            double[3] df_efficiency
            unsigned char[3][256] dbuffer
            long long[3] dc
        assert self.decompression_pos > 0, "The buffer to save is empty."
        while dp2 < decomp_len:
            if cp >= self.compression_pos:
                raise Exception("compression running out of space, try enlarging the compression buffer.")
            retest = 1
            cur_val = decomp2[dp2]
            cur_dp2 = dp2 + 1
            while cur_dp2 < decomp_len and retest < REPEAT_MAX_LEN and decomp2[cur_dp2] == cur_val:
                retest += 1
                cur_dp2 += 1
            re_efficiency = retest / 3.

            if re_efficiency < MAX_EFF[0]:
                df_efficiency[0] = df_efficiency[1] = df_efficiency[2] = 0.
                dc[0] = dc[1] = dc[2] = 0
                if dp2 > 0:
                    for i in range(3):
                        prior_val = decomp2[dp2 - 1]
                        cur_dp2 = dp2
                        for cur_dp2 in range(dp2, dp2 + min(decomp_len - dp2, MAX_LEN[i])):
                            delta = decomp2[cur_dp2] - prior_val
                            if delta > ran[i][1] or delta < ran[i][0]:
                                break
                            prior_val = decomp2[cur_dp2]
                            dbuffer[i][cur_dp2 - dp2] = ran[i][1] - delta if delta < 0 else delta
                        else:
                            cur_dp2 += 1
                        df_efficiency[i] = (cur_dp2 - dp2) / ((cur_dp2 - dp2) * (3 + i) / 16. + 1)
                        dc[i] = cur_dp2
                        if not self.pbd16_full_blood:
                            break
                    else:
                        if df_efficiency[1] > df_efficiency[0]:
                            if df_efficiency[2] > df_efficiency[1]:
                                i = 2
                            else:
                                i = 1
                        elif df_efficiency[2] > df_efficiency[0]:
                            i = 2
            if re_efficiency >= MAX_EFF[0] or re_efficiency > df_efficiency[i] and re_efficiency > 1.:
                comp[cp] = retest + 222
                cp += 1
                pcp2 = <unsigned short*>&comp[cp]
                pcp2[0] = cur_val
                cp += 2
                dp2 += retest
                active_literal_index = -1
            elif df_efficiency[i] > 1.:
                comp[cp] = dc[i] - dp2 + gap[i]
                cp += 1
                carry = 0
                shift = 8
                for cur_dp2 in range(dc[i] - dp2):
                    shift -= shift_bits[i]
                    if shift > 0:
                        carry |= dbuffer[i][cur_dp2] << shift
                    else:
                        carry |= dbuffer[i][cur_dp2] >> -shift
                        comp[cp] = carry
                        cp += 1
                        shift += 8
                        carry = dbuffer[i][cur_dp2] << shift
                else:
                    if shift != 8:
                        comp[cp] = carry
                        cp += 1
                active_literal_index = -1
                dp2 = dc[i]
            else:
                if active_literal_index < 0 or comp[active_literal_index] >= 31:
                    comp[cp] = 0
                    active_literal_index = cp
                    cp += 1
                else:
                    comp[active_literal_index] += 1
                pcp2 = <unsigned short*>&comp[cp]
                pcp2[0] = cur_val
                cp += 2
                dp2 += 1
        return cp

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef void save(self, path: str | os.PathLike, np.ndarray img):
        """
        :param path: output image path of v3dpbd.
        :param img: 4D numpy array (C,Z,Y,X) of either uint8 or uint16.
        """
        assert img.ndim == 4, "The image has to be 4D"
        assert img.dtype in [np.uint8, np.uint16], "The pixel type has to be uint8 or uint16"
        cdef:
            bytearray header
            int[4] sz = [img.shape[0], img.shape[1], img.shape[2], img.shape[3]]
            int[:] size = sz
            const unsigned char[:] p = str(path).encode('utf-8')
            long long compression_size, channel_len
            short datatype
            FILE * f = fopen(<const char *> &p[0], <char *> 'wb')
        if f is NULL:
            raise Exception("Fail to open file for writing.")
        try:
            endian = '<' if self.endian_sys == LITTLE else '>'
            if img.dtype == np.uint8:
                datatype = 1
            elif img.dtype == np.uint16:
                datatype = 2
            else:
                raise Exception("Unsupported datatype.")
            header = bytearray(FORMAT_KEY + self.endian_sys + struct.pack(f'{endian}hiiii', datatype, *size[::-1]))
            assert fwrite(PyByteArray_AsString(header), HEADER_SIZE, 1, f) == 1, "Header writing failed."
            channel_len = sz[0] * sz[1] * sz[2]
            self.compression_pos = channel_len * sz[3] * datatype * COMPRESSION_ENLARGEMENT
            self.compression_buffer = bytearray(self.compression_pos)
            if datatype == 2 and img.dtype.byteorder not in ['=', endian]:
                img = img.byteswap()
            self.decompression_buffer = bytearray(img.tobytes())
            self.decompression_pos = len(self.decompression_buffer)
            if datatype == 1:
                compression_size = self.compress_pbd8()
            elif datatype == 2:
                compression_size = self.compress_pbd16()
            else:
                raise Exception("Invalid datatype.")
            assert fwrite(<void*>PyByteArray_AsString(self.compression_buffer), compression_size, 1, f) == 1, \
                "Buffer saving failed."
        finally:
            fclose(f)



import unittest
from v3dpy.terafly import TeraflyInterface
from pathlib import Path
import numpy as np


test_data_path = Path(r'D:\GitHub\TeraQC\data')
outdir = Path('testoutput')


class TeraflyTest(unittest.TestCase):

    def test_image_dims(self):
        for tfpath in test_data_path.rglob('RES*'):
            y, x, z = [int(i) for i in tfpath.name[4:-1].split('x')]
            t = TeraflyInterface(tfpath)
            self.assertTupleEqual(t.get_dim(), (x, y, z, 1), "Dims are not equal")

    def test_image_crop(self):
        for tfpath in test_data_path.rglob('RES*'):
            t = TeraflyInterface(tfpath)

            # center block
            size = np.array(t.get_dim()[:3])
            half_block_size = np.array([128, 128, 64]) // 2
            start = size // 2 - half_block_size
            end = size // 2 + half_block_size - 1

            img = t.get_sub_volume(start[0], end[0], start[1], end[1], start[2], end[2])
            print(img.max())
            # PBD(pbd16_full_blood=False).save(outdir / f'{tfpath.parent.name}.v3dpbd', img)


if __name__ == '__main__':
    unittest.main()



import io
import os
import struct
import sys
import cython
cimport cython
import numpy as np
cimport numpy as np


DEF FORMAT_KEY_4 = b"raw_image_stack_by_hpeng"
DEF FORMAT_KEY_5 = b"raw5image_stack_by_hpeng"
DEF FORMAT_LEN = 24


cdef class Raw:
    """
    For vaa3d raw formats, allowing uint8, uint16 and float32 data types. The image dimension is presumed to be 4.

    As a raw format, it not only stores the image buffer, but also the image dimension sizes and its data type in the
    header. It also stores the endian of the buffer.

    This interface saves and loads a multi-dimension numpy array of either of the 3 data types. It also compares the endian of
    the file/numpy array and the machine to make it work properly. Based on the v3draw format key in the header, the
    array can be 4D or 5D.

    modified from v3d_external/v3d_main/basic_c_fun/stackutils.cpp

    by Zuohan Zhao

    2022/5/8
    """
    cdef:
        bint sz2byte

    def __init__(self, sz2byte = False):
        """

        :param sz2byte: set size array to be 2 byte (short int), for compatibility. Default as False.
        """
        self.sz2byte = sz2byte

    cpdef np.ndarray load(self, path: str | os.PathLike, int choose = -1):
        """
        :param path: input image path of v3draw.
        :param choose: choose a channel(4D) or stack(5D) to load, starting from 0, default as -1, meaning all.
        :return: a numpy array of 4D or 5D based on the format key.
        """
        cdef:
            short datatype
            list sz
            bytes endian_code_data
            str endian, dt
            char dim, header_sz, i
            bytes format_key
            long long bulk_sz, filesize

        filesize = os.path.getsize(path)
        assert filesize >= FORMAT_LEN, "File size too small, file might be corrupted"

        with open(path, "rb") as f:
            format_key = f.read(FORMAT_LEN)
            if format_key == FORMAT_KEY_4:
                dim = 4
            elif format_key == FORMAT_KEY_5:
                dim = 5
            else:
                raise RuntimeError("Format key isn't for v3draw")
            if self.sz2byte:
                header_sz = FORMAT_LEN + dim * 2 + 2 + 1
            else:
                header_sz = FORMAT_LEN + dim * 4 + 2 + 1
            assert filesize >= header_sz, "File size too small, file might be corrupted"
            endian_code_data = f.read(1)
            if endian_code_data == b'B':
                endian = '>'
            elif endian_code_data == b'L':
                endian = '<'
            else:
                raise RuntimeError('Endian code should be either B/L')
            datatype = struct.unpack(f'{endian}h', f.read(2))[0]
            if datatype == 1:
                dt = 'u1'
            elif datatype == 2:
                dt = 'u2'
            elif datatype == 4:
                dt = 'f4'
            else:
                raise RuntimeError('v3draw data type can only be 1/2/4')
            if self.sz2byte:
                sz = list(struct.unpack(f'{endian}{dim}h', f.read(dim * 2)))
            else:
                sz = list(struct.unpack(f'{endian}{dim}i', f.read(dim * 4)))
            bulk_sz = 1
            for i in range(dim - 1):
                bulk_sz *= sz[i]
            assert bulk_sz * sz[-1] * datatype + header_sz == filesize, "file size doesn't match with the image"
            if choose < 0:
                return np.frombuffer(f.read(), endian + dt).reshape(sz[::-1])
            else:
                assert choose < sz[-1], "Choose index exceeding the range"
                f.seek(bulk_sz * datatype * choose, 1)
                img = np.frombuffer(f.read(bulk_sz * datatype), endian + dt)
                return img.reshape(sz[-2::-1])

    cpdef void save(self, path: str | os.PathLike, np.ndarray img):
        """
        :param path: output image path of v3draw.
        :param img: the image array to save. 
        """
        assert img.ndim == 4, "The image has to be 4D"
        assert img.dtype in [np.uint8, np.uint16], "The pixel type has to be uint8 or uint16"
        cdef:
            short datatype
            list sz
            bytes endian_code_data, format
            str endian, bo = img.dtype.byteorder
            bytes header
            char dim = img.ndim

        if dim == 4:
            format = FORMAT_KEY_4
        elif dim == 5:
            format = FORMAT_KEY_5
        else:
            raise RuntimeError("Dimension not supported by v3draw")

        sz = [img.shape[i] for i in range(dim)]
        sz.extend([1] * (dim - len(sz)))
        sz.reverse()

        if bo == '>' or bo == '=' and sys.byteorder == 'big':
            endian_code_data = b'B'
            endian = '>'
        else:
            endian_code_data = b'L'
            endian = '<'

        if img.dtype == np.uint8:
            datatype = 1
        elif img.dtype == np.uint16:
            datatype = 2
        elif img.dtype == np.float32:
            datatype = 4
        else:
            raise RuntimeError("numpy data type not supported by v3draw")

        header = struct.pack(f'{endian}{FORMAT_LEN}sch4{"h" if self.sz2byte else "i"}',
                             format, endian_code_data, datatype, *sz)

        with open(path, 'wb') as f:
            f.write(header)
            f.write(img.tobytes())




MDATA_BIN_FILE_NAME = "mdata.bin"  # name of binary metadata file
MDATA_BIN_FILE_VERSION = 2  # version of binary metadata file
MC_MDATA_BIN_FILE_NAME = "cmap.bin"  # name of binary metadata file for multichannel volumes
FORMAT_MDATA_FILE_NAME = ".iim.format"  # name of format metadata file
CHANNEL_PREFIX = "CH_"  # prefix identifying a folder containing data of a certain channel
TIME_FRAME_PREFIX = "T_"  # prefix identifying a folder containing data of a certain time frame
DEF_IMG_DEPTH = 8  # default image depth
NUL_IMG_DEPTH = 0  # invalid image depth
NATIVE_RTYPE = 0  # loadVolume returns the same bytes per channel as in the input image
DEF_IMG_FORMAT = "tif"  # default image format
STATIC_STRINGS_SIZE = 1024  # size of static C-strings
RAW_FORMAT = "Vaa3D raw"  # unique ID for the RawVolume class
SIMPLE_RAW_FORMAT = "Vaa3D raw (series, 2D)"  # unique ID for the SimpleVolumeRaw class
STACKED_RAW_FORMAT = "Vaa3D raw (tiled, 2D)"  # unique ID for the StackedVolume class
TILED_FORMAT = "Vaa3D raw (tiled, 3D)"  # unique ID for the TiledVolume class
TILED_MC_FORMAT = "Vaa3D raw (tiled, 4D)"  # unique ID for the TiledMCVolume class
TIF3D_FORMAT = "TIFF (3D)"  # unique ID for multipage TIFF format (nontiled)
SIMPLE_FORMAT = "TIFF (series, 2D)"  # unique ID for the SimpleVolume class
STACKED_FORMAT = "TIFF (tiled, 2D)"  # unique ID for the StackedVolume class
TILED_TIF3D_FORMAT = "TIFF (tiled, 3D)"  # unique ID for multipage TIFF format (tiled)
TILED_MC_TIF3D_FORMAT = "TIFF (tiled, 4D)"  # unique ID for multipage TIFF format (nontiled, 4D)
UNST_TIF3D_FORMAT = "TIFF (unstitched, 3D)"  # unique ID for multipage TIFF format (nontiled, 4D)
BDV_HDF5_FORMAT = "HDF5 (BigDataViewer)"  # unique ID for BDV HDF5
IMS_HDF5_FORMAT = "HDF5 (Imaris IMS)"  # unique ID for IMS HDF5
MAPPED_FORMAT = "Mapped Volume"  # unique ID for mapped volumes
MULTISLICE_FORMAT = "MultiSlice Volume"  # unique ID for multi-slice volumes
MULTICYCLE_FORMAT = "MultiCycle Volume"  # unique ID for multi-cycle volumes
VOLATILE_FORMAT = "Volatile Volume"  # unique ID for volatile volumes
TIME_SERIES = "Time series"  # unique ID for the TimeSeries classes




import unittest
from v3dpy.loaders.pbd import PBD
from v3dpy.loaders.raw import Raw
from pathlib import Path
import numpy as np

test_data_path = Path('d:/')


class LoaderTest(unittest.TestCase):

    def test_raw8(self):
        loader = Raw()
        in_img = loader.load(test_data_path / '8.v3draw')
        loader.save(test_data_path / '8_.v3draw', in_img)
        out_img = loader.load(test_data_path / '8_.v3draw')
        self.assertEqual((in_img == out_img).all(), True)  # add assertion here

    def test_raw16(self):
        loader = Raw()
        in_img = loader.load(test_data_path / '16.v3draw')
        loader.save(test_data_path / '16_.v3draw', in_img)
        out_img = loader.load(test_data_path / '16_.v3draw')
        self.assertEqual((in_img == out_img).all(), True)  # add assertion here

    def test_pbd8(self):
        loader = PBD()
        in_img = loader.load(test_data_path / '8.v3dpbd')
        loader.save(test_data_path/ '8_.v3dpbd', in_img)
        out_img = loader.load(test_data_path / '8_.v3dpbd')
        self.assertEqual((in_img == out_img).all(), True)  # add assertion here

    def test_pbd16_dbg(self):
        loader = PBD(pbd16_full_blood=False)
        in_img = loader.load(test_data_path / '16.v3dpbd')
        loader.save(test_data_path / '16_.v3dpbd', in_img)
        out_img = loader.load(test_data_path / '16_.v3dpbd')
        self.assertEqual((in_img == out_img).all(), True)  # add assertion here

    def test_pbd16_fb(self):
        loader = PBD(pbd16_full_blood=True)
        in_img = loader.load(test_data_path / '16.v3dpbd')
        loader.save(test_data_path / '16_.v3dpbd', in_img)
        out_img = loader.load(test_data_path / '16_.v3dpbd')
        self.assertEqual((in_img == out_img).all(), True)  # add assertion here

    def test_pbd16_new(self):
        loader = PBD()
        in_img = loader.load(test_data_path / '16.v3dpbd')
        loader.save16_halving(test_data_path / '16_.v3dpbd', in_img)
        # out_img = loader.load(test_data_path / 'pbd16.v3dpbd')
        # self.assertEqual((in_img == out_img).all(), True)  # add assertion here


if __name__ == '__main__':
    unittest.main()



import unittest
from v3dpy.loaders import PBD
from v3dpy.neuron_utilities import swc_handler
from v3dpy.neuron_utilities import radius
# import numpy as np

#
# def neuron_radius(tree, img, is2d, bkg_thr):
#     assert img.ndim == 3
#     tree = [list(t) for t in tree]
#     img_ = img.astype(np.float32)
#     for t in tree:
#         x, y, z = t[2:5]
#         t[5] = marker_radius_hanchuan_xy(x, y, z, img_, bkg_thr)
#     return [tuple(t) for t in tree]
#
#
# def marker_radius_hanchuan_xy(x, y, z, img, thr):
#     sz0 = img.shape[2]
#     sz1 = img.shape[1]
#     k = int(z)
#     max_r = sz0 / 2
#     max_r = min(max_r, sz1 / 2)
#     for ir in range(1, int(max_r + 1)):
#         total_num = background_num = 0
#
#         for dy in range(-ir, (ir + 1)):
#             for dx in range(-ir, (ir + 1)):
#                 total_num += 1
#                 r = (dx*dx + dy*dy) **.5
#                 if ir - 1 < r <= ir:
#                     i = int(x + dx)
#                     if i < 0 or i >= sz0:
#                         return ir
#                     j = int(y + dy)
#                     if j < 0 or j >= sz1:
#                         return ir
#                     if img[k, j, i] <= thr:
#                         background_num += 1
#                         if background_num / total_num > 0.001:
#                             return ir
#     return 1


class MyTestCase(unittest.TestCase):
    def test_something(self):
        swc = r"D:\rectify\manual\15257_13263_25518_2294.swc"
        img = r"D:\rectify\my\15257_13263_25518_2294.v3dpbd"
        img = PBD().load(img)[0]
        swc = swc_handler.parse_swc(swc)
        res = radius.neuron_radius(swc, img, True, 3)
        swc_handler.write_swc(res, 'profiled.swc')


if __name__ == '__main__':
    unittest.main()



"""
.. include:: ../README.md
"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("v3d-py-helper")
except PackageNotFoundError:
    __version__ = "UNKNOWN"



"""
Loaders for TeraFly formats. To use, first initialize an object with the TeraConverted directory path.
Two functions are offered for checking the dims and retrieving the sub blocks.
The indexing style of TeraFly is [start, end]
"""

import os

from .volume_managers import VirtualVolume, TiledVolume
from pathlib import Path
from .config import *


__all__ = ['TeraflyInterface']


class TeraflyInterface:
    """
    Currently only support 3D tiff tiles.
    """
    def __init__(self, path: os.PathLike | str):
        self._path = Path(path)
        self._volume: VirtualVolume
        self.update_metadata()

    def update_metadata(self):
        """
        now it only supports TiledVolume (Tiled Tiff/V3DRaw 3D)
        :param path: teraconverted brain / resolution
        :return: an instance of the image manager from the path
        """
        self._volume = None
        if self._path.is_dir():
            format = "unknown"
            if (self._path / FORMAT_MDATA_FILE_NAME).is_file():
                try:
                    with open(self._path / FORMAT_MDATA_FILE_NAME) as f:
                        format = f.readline().rstrip()
                    if format == TILED_MC_FORMAT:
                        raise NotImplementedError
                    elif format == STACKED_FORMAT:
                        raise NotImplementedError
                    elif format == TILED_FORMAT:
                        self._volume = TiledVolume(self._path)
                    elif format == SIMPLE_FORMAT:
                        raise NotImplementedError
                    elif format == SIMPLE_RAW_FORMAT:
                        raise NotImplementedError
                    elif format == TIME_SERIES:
                        raise NotImplementedError
                    else:
                        raise ValueError(f"Cannot recognize format {format}")
                except:
                    raise IOError(f"Cannot import {format} at {self._path}")
            if self._volume is None:
                # try:
                #     self._volume = TiledMCVolume(path)
                # except:
                #     print(f"cannot import TiledMCVolume at {path}")
                #     try:
                #         self._volume = StackedVolume(path)
                #     except:
                #         print(f"cannot import StackedVolume at {path}")
                try:
                    self._volume = TiledVolume(self._path)
                except:
                    print(f"Cannot import TiledVolume at {self._path}")
                    # try:
                    #     self._volume = SimpleVolume(path)
                    # except:
                    #     print(f"cannot import SimpleVolume at {path}")
                    #     try:
                    #         self._volume = SimpleVolumeRaw(path)
                    #     except:
                    #         print(f"cannot import SimpleVolumeRaw at {path}")
                    #         try:
                    #             self._volume = TimeSeries(path)
                    #         except:
                    #             print(f"cannot import TimeSeries at {path}")
                    raise f"Generic error occurred when importing self._volume at {self._path}"
        elif self._path.is_file():
            if self._path.suffix in ['.raw', '.v3draw', '.RAW', '.V3DRAW', '.tif', '.tiff', '.TIF', '.TIFF']:
                raise NotImplementedError
            elif self._path.suffix in ['.xml', '.XML']:
                raise NotImplementedError
            else:
                raise ValueError(f"Unsupported file extensions for {self._path}")
        else:
            raise ValueError(f"Path {self._path} does not exist")

    def get_dim(self) -> tuple[int, int, int, int]:
        """
        Get TeraConvert dimensionality range.
        :return: (max x, max y, max z, max c)
        """
        return self._volume.DIM_H, self._volume.DIM_V, self._volume.DIM_D, self._volume.DIM_C

    def get_sub_volume(self, x0: int, x1: int, y0: int, y1: int, z0: int, z1: int):
        """
        Different from Vaa3D, it returns the image of its original pixel type.
        The indexing is pixel-wise, so you have to use different coordinates for different resolutions
        If the loaded image is tif, it will be properly flipped

        :param path: teraconverted brain / resolution
        :param x0: starting x
        :param x1: ending x
        :param y0: starting y
        :param y1: ending y
        :param z0: starting z
        :param z1: ending z
        :return: the image crop
        """
        return self._volume.load_sub_volume(y0, y1, x0, x1, z0, z1)



from libc.stdint cimport uint32_t, uint16_t, uint8_t, int64_t, uint64_t
from libc.string cimport memcpy
from libc.math cimport floor, ceil
import cython
cimport numpy as cnp
import numpy as np


cdef extern from "tiffio.h":
    ctypedef struct TIFF
    ctypedef int64_t ttile_t
    ctypedef uint64_t tstrip_t
    ctypedef uint64_t tsize_t
    ctypedef void * tdata_t

    TIFF * TIFFOpen(const char * name, const char * mode)
    void TIFFClose(TIFF * tif)

    void TIFFSetWarningHandler(void * handler)
    void TIFFSetErrorHandler(void * handler)

    int TIFFGetField(TIFF * tif, unsigned int tag, ...)
    int TIFFReadDirectory(TIFF * tif)
    int TIFFIsTiled(TIFF * tif)
    tsize_t TIFFTileSize(TIFF * tif)
    ttile_t TIFFComputeTile(TIFF * tif, uint32_t x, uint32_t y, uint32_t z, uint32_t s)
    tsize_t TIFFReadEncodedTile(TIFF * tif, ttile_t tile, tdata_t buf, tsize_t size)
    tstrip_t TIFFNumberOfStrips(TIFF * tif)
    tsize_t TIFFStripSize(TIFF * tif)
    tsize_t TIFFReadEncodedStrip(TIFF * tif, tstrip_t strip, tdata_t buf, tsize_t size)
    tsize_t TIFFReadRawStrip(TIFF * tif, tstrip_t strip, tdata_t buf, tsize_t size)
    int TIFFIsByteSwapped(TIFF * tif)
    int TIFFSetDirectory(TIFF * tif, uint16_t dirnum)
    ttile_t TIFFNumberOfTiles(TIFF * tif)

    enum:  # TIFF tags
        TIFFTAG_IMAGEWIDTH
        TIFFTAG_IMAGELENGTH
        TIFFTAG_BITSPERSAMPLE
        TIFFTAG_SAMPLESPERPIXEL
        TIFFTAG_PAGENUMBER
        TIFFTAG_PHOTOMETRIC
        TIFFTAG_COMPRESSION
        TIFFTAG_PLANARCONFIG
        TIFFTAG_TILEWIDTH
        TIFFTAG_TILELENGTH
        TIFFTAG_TILEDEPTH
        TIFFTAG_ROWSPERSTRIP


cdef void swap2bytes(void* targetp):
    cdef unsigned char* tp = <unsigned char*>targetp
    cdef unsigned char a = tp[0]
    tp[0] = tp[1]
    tp[1] = a

cdef void swap4bytes(void* targetp):
    cdef unsigned char* tp = <unsigned char*>targetp
    cdef unsigned char a = tp[0]
    tp[0] = tp[3]
    tp[3] = a
    a = tp[1]
    tp[1] = tp[2]
    tp[2] = a


cdef void close_tiff3d_file(void* fhandle):
    TIFFClose(<TIFF*> fhandle)


cdef void* load_tiff3d2metadata(const char* filename, unsigned int& sz0, unsigned int& sz1, unsigned int& sz2,
                                unsigned int& sz3, int& datatype, bint& b_swap):
    cdef:
        uint32_t XSIZE
        uint32_t YSIZE
        uint16_t bpp
        uint16_t spp
        uint16_t Cpage
        uint16_t Npages
        TIFF* input
        int check
        char* mode = 'r'

    # disable warning and error handlers to avoid messages on unrecognized tags
    TIFFSetWarningHandler(NULL)
    TIFFSetErrorHandler(NULL)

    input = TIFFOpen(filename, mode)
    if not input:
        raise IOError("Cannot open the file.")

    check = TIFFGetField(input, TIFFTAG_IMAGEWIDTH, &XSIZE)
    if not check:
        TIFFClose(input)
        raise IOError("Image width of undefined.")

    check = TIFFGetField(input, TIFFTAG_IMAGELENGTH, &YSIZE)
    if not check:
        TIFFClose(input)
        raise IOError("Image length of undefined.")

    check = TIFFGetField(input, TIFFTAG_BITSPERSAMPLE, &bpp)
    if not check:
        TIFFClose(input)
        raise IOError("Undefined bits per sample.")

    check = TIFFGetField(input, TIFFTAG_SAMPLESPERPIXEL, &spp)
    if not check:
        spp = 1

    check = TIFFGetField(input, TIFFTAG_PAGENUMBER, &Cpage, &Npages)
    if check != 1 or Npages == 0:
        Npages = 0
        while TIFFReadDirectory(input):
            Npages += 1

    sz0 = XSIZE
    sz1 = YSIZE
    sz2 = Npages
    sz3 = spp
    datatype = bpp // 8

    b_swap = TIFFIsByteSwapped(input)
    return <void*> input


cdef void copydata(unsigned char *psrc, uint32_t stride_src, unsigned char *pdst, uint32_t stride_dst, uint32_t width, uint32_t len):
    cdef uint32_t i
    for i in range(len):
        memcpy(pdst, psrc, width * sizeof(unsigned char))
        psrc += stride_src
        pdst += stride_dst


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void read_tiff_3d_file_to_buffer(void* fhandler,
                                     unsigned char * img,
                                     unsigned int img_width,
                                     unsigned int img_height,
                                     unsigned int first,
                                     unsigned int last,
                                     bint b_swap,
                                     int downsampling_factor,
                                     int start_i,
                                     int end_i,
                                     int start_j,
                                     int end_j):
    cdef:
        uint32_t rps
        uint16_t spp, bpp, orientation, photo, comp, planar_config, c
        int check, strips_per_image, last_strip_size
        uint32_t XSIZE
        uint32_t YSIZE
        TIFF* input = <TIFF*>fhandler

    check = TIFFGetField(input, TIFFTAG_BITSPERSAMPLE, &bpp)
    if not check:
        raise IOError("Undefined bits per sample.")

    check = TIFFGetField(input, TIFFTAG_SAMPLESPERPIXEL, &spp)
    if not check:
        raise IOError("Undefined samples per sample.")

    check = TIFFGetField(input, TIFFTAG_PHOTOMETRIC, &photo)
    if not check:
        raise IOError("Cannot determine photometric interpretation.")

    check = TIFFGetField(input, TIFFTAG_COMPRESSION, &comp)
    if not check:
        raise IOError("Cannot determine compression technique.")

    check = TIFFGetField(input, TIFFTAG_PLANARCONFIG, &planar_config)
    if not check:
        raise IOError("Cannot determine planar configuration.")

    start_i = 0 if start_i == -1 else start_i
    end_i = img_height - 1 if end_i == -1 else end_i
    start_j = 0 if start_j == -1 else start_j
    end_j = img_width - 1 if end_j == -1 else end_j

    cdef:
        uint32_t tilewidth, tilelength, tiledepth
        tsize_t tilenum, tilesize, tilenum_width, tilenum_length
        ttile_t tile
        cnp.ndarray[cnp.uint8_t, ndim=1] npdata
        tdata_t data
        unsigned char *psrc  # pointer in the tile buffer to the top left pixel of the current block to be copied
        unsigned char *pdst  # pointer in the image buffer to the top left pixel of the current block to be filled
        uint32_t stride_src, stride_dst
        tsize_t i  # row index in the slice of the top left pixel of the current block to be filled
        tsize_t j  # column index in the slice of the top left pixel of the current block to be filled
        uint32_t width  # width of the current block to be filled (in pixels)
        uint32_t len  # length of the current block to be filled (in pixels)
        int page = 0

    check = TIFFIsTiled(input)
    if check:
        # checks
        if TIFFGetField(input, TIFFTAG_TILEDEPTH, &tiledepth):
            raise IOError("Tiling among slices (z direction) not supported.")
        if spp > 1:
            if TIFFGetField(input, TIFFTAG_PLANARCONFIG, &planar_config):
                if planar_config > 1:
                    raise IOError("Non-interleaved multiple channels not supported with tiling.")

        # tiling is in x,y only
        TIFFGetField(input, TIFFTAG_TILEWIDTH, &tilewidth)
        TIFFGetField(input, TIFFTAG_TILELENGTH, &tilelength)
        tilenum = TIFFNumberOfTiles(input)
        tilesize = TIFFTileSize(input)
        tilenum_width = (img_width % tilewidth) if (img_width / tilewidth) + 1 else img_width / tilewidth
        tilenum_length = (img_height % tilelength) if (img_height / tilelength) + 1 else img_height / tilelength

        npdata = np.zeros(tilesize, dtype=np.uint8)
        data = <unsigned char *> npdata.data
        stride_src = tilewidth * spp  # width of tile (in bytes)
        stride_dst = (end_j - start_j + 1) * spp  # width of subregion (in bytes)

        while True:
            # Calculate initial parameters
            psrc = <unsigned char *> data + ((start_i % tilelength) * tilewidth + (start_j % tilewidth)) * spp
            pdst = img
            len = tilelength - (start_i % tilelength)
            tile = TIFFComputeTile(input, start_j, start_i, 0, 0)

            # Loop over all tiles in the TIFF file
            i = start_i
            while i <= end_i:
                width = tilewidth - (start_j % tilewidth)
                j = start_j
                while j <= end_j:
                    # Read the current tile into the data buffer
                    TIFFReadEncodedTile(input, tile, data, <tsize_t> -1)

                    # Copy the current block from the tile buffer to the image buffer
                    copydata(psrc, stride_src, pdst, stride_dst, width * spp, len)

                    j += width
                    tile += 1
                    psrc = <unsigned char *> data + ((i % tilelength) * tilewidth) * spp
                    pdst += width * spp
                    width = tilewidth if ((tile % tilenum_width + 1) * tilewidth <= end_j + 1) else (end_j + 1) % tilewidth

                i += len
                tile = TIFFComputeTile(input, start_j, i, 0, 0)
                psrc = <unsigned char *> data + ((i % tilelength) * tilewidth + (start_j % tilewidth)) * spp
                pdst = img + (i - start_i) * stride_dst
                len = tilelength if ((tile / tilenum_width + 1) * tilelength <= end_i + 1) else (end_i + 1) % tilelength

                page += 1
            if not(<unsigned int>page < (last - first + 1) and TIFFReadDirectory(input)):
                break
        return

    if not TIFFGetField(input, TIFFTAG_ROWSPERSTRIP, &rps):
        rps = img_height
    strips_per_image = (img_height + rps - 1) / rps
    last_strip_size = img_height % rps
    if last_strip_size == 0:
        last_strip_size = rps

    cdef unsigned char * buf = img
    cdef cnp.ndarray[cnp.uint8_t, ndim=1] nprowbuf
    cdef unsigned char * rowbuf
    cdef unsigned char * bufptr
    cdef int strip_index

    if downsampling_factor == 1:  # read without downsampling
        if start_i < 0 or end_i >= <int>img_height or start_j < 0 or end_j >= <int>img_width or \
                start_i >= end_i or start_j >= end_j:
            raise IOError("Wrong substack indices.")

        if start_i == 0 and end_i == <int>img_height - 1 and start_j == 0 and end_j == <int>img_width - 1:  # read whole images from files
            if not TIFFSetDirectory(input, first):
                raise IOError("Cannot open the requested first strip.")

            while True:
                for i in range(strips_per_image - 1):
                    if comp == 1:
                        TIFFReadRawStrip(input, i, buf, spp * rps * img_width * (bpp / 8))
                        buf += spp * rps * img_width * (bpp / 8)
                    else:
                        TIFFReadEncodedStrip(input, i, buf, spp * rps * img_width * (bpp / 8))
                        buf += spp * rps * img_width * (bpp / 8)

                if comp == 1:
                    TIFFReadRawStrip(input, strips_per_image - 1, buf, spp * last_strip_size * img_width * (bpp / 8))
                else:
                    TIFFReadEncodedStrip(input, strips_per_image - 1, buf, spp * last_strip_size * img_width * (bpp / 8))
                buf += spp * last_strip_size * img_width * (bpp / 8)

                page += 1
                if not (<unsigned int>page < last - first + 1 and TIFFReadDirectory(input)):
                    break
        else:  # read only a subregion of images from files
            if not TIFFGetField(input, TIFFTAG_IMAGEWIDTH, &XSIZE):
                raise IOError("Image width of undefined.")
            if not TIFFGetField(input, TIFFTAG_IMAGELENGTH, &YSIZE):
                raise IOError("Image length of undefined.")

            nprowbuf = np.zeros(spp * rps * XSIZE * (bpp/8), dtype=np.uint8)
            rowbuf = <uint8_t *> nprowbuf.data

            while True:
                if not TIFFSetDirectory(input, first + page):
                    raise IOError("Cannot open next requested strip.")

                strip_index = (start_i / rps) - 1  # the strip preceeding the first one
                for i in range(start_i, end_i + 1):
                    if floor(i / rps) > strip_index:  # read a new strip
                        strip_index = <int>floor(i / rps)
                        if comp == 1:
                            TIFFReadRawStrip(input, strip_index, rowbuf,
                                             spp * (rps if strip_index < strips_per_image else last_strip_size) * XSIZE * (bpp / 8))
                        else:
                            TIFFReadEncodedStrip(input, strip_index, rowbuf,
                                                 spp * (rps if strip_index < strips_per_image else last_strip_size) * XSIZE * (bpp / 8))

                    bufptr = rowbuf + <uint32_t>((i % rps) * (spp * XSIZE * (bpp / 8)))
                    if bpp == 8:
                        for j in range(end_j - start_j + 1):
                            for c in range(spp):
                                buf[j * spp + c] = bufptr[j * spp + c]
                    else:
                        for j in range(end_j - start_j + 1):
                            for c in range(spp):
                                (<uint16_t *> buf)[j * spp + c] = (<uint16_t *> bufptr)[j * spp + c]
                    buf += spp * (end_j - start_j + 1) * (bpp / 8)

                page += 1
                if not (<unsigned int>page < last - first + 1):
                    break
    else:  # read with downsampling
        # preliminary checks
        if start_i != 0 or end_i != <int>img_height - 1 or start_j != 0 or end_j != <int>img_width - 1:  # a subregion has been requested
            raise IOError("Subregion extraction not supported with downsampling.")
        if not TIFFGetField(input, TIFFTAG_IMAGEWIDTH, &XSIZE):
            raise IOError("Image width of undefined.")
        if not TIFFGetField(input, TIFFTAG_IMAGELENGTH, &YSIZE):
            raise IOError("Image length of undefined.")

        if ceil(XSIZE / downsampling_factor) < img_width:
            raise IOError("Requested image width too large.")
        if ceil(YSIZE / downsampling_factor) < img_height:
            raise IOError("Requested image height too large.")

        nprowbuf = np.zeros(spp * rps * XSIZE * (bpp / 8), dtype=np.uint8)
        rowbuf = <uint8_t *> nprowbuf.data

        while True:
            if not TIFFSetDirectory(input, (first + page) * downsampling_factor):
                raise IOError("Cannot open next requested strip.")

            strip_index = -1  # the strip preceeding the first one
            for i in range(img_height):
                if floor(i * downsampling_factor / rps) > strip_index:  # read a new strip
                    strip_index = int(floor(i * downsampling_factor / rps))
                    if comp == 1:
                        TIFFReadRawStrip(input, strip_index, rowbuf,
                                         spp * (rps if strip_index < strips_per_image else last_strip_size) * XSIZE * (bpp / 8))
                    else:
                        TIFFReadEncodedStrip(input, strip_index, rowbuf,
                                             spp * (rps if strip_index < strips_per_image else last_strip_size) * XSIZE * (bpp / 8))

                bufptr = rowbuf + <uint32_t>(((i * downsampling_factor) % rps) * (spp * XSIZE * (bpp / 8)))
                if bpp == 8:
                    for j in range(img_width):
                        for c in range(spp):
                            buf[j * spp + c] = bufptr[j * spp * downsampling_factor + c]
                else:
                    for j in range(img_width):
                        for c in range(spp):
                            (<uint16_t *> buf)[j * spp + c] = (<uint16_t *> bufptr)[j * spp * downsampling_factor + c]
                buf += spp * img_width * (bpp / 8)

            page += 1

            if not (<unsigned int>page < last - first + 1):
                break

    cdef tsize_t total = img_width * img_height * spp * (last-first+1)
    if b_swap:
        if bpp / 8 == 2:
            for i in range(total):
                swap2bytes(<void *> (img + 2 * i))
        elif bpp / 8 == 4:
            for i in range(total):
                swap4bytes(<void *> (img + 4 * i))


cdef public void read_tiff_3d_file_to_buffer(void* fhandler,
                                             unsigned char * img,
                                             unsigned int img_width,
                                             unsigned int img_height,
                                             unsigned int first,
                                             unsigned int last,
                                             bint b_swap,
                                             int downsampling_factor,
                                             int start_i,
                                             int end_i,
                                             int start_j,
                                             int end_j)

cdef public void close_tiff3d_file(void* fhandle)

cdef public void* load_tiff3d2metadata(const char* filename, unsigned int& sz0, unsigned int& sz1, unsigned int& sz2,
                                       unsigned int& sz3, int& datatype, bint& b_swap)


# Import relevant Cython modules
import numpy as np
cimport cython
from .tiff_manage cimport read_tiff_3d_file_to_buffer, load_tiff3d2metadata, close_tiff3d_file
from libc.stdint cimport uint8_t, int64_t, uint16_t, int32_t, uint32_t
from libc.string cimport memcpy
cimport numpy as cnp


cdef class VirtualFmtMngr:
    cdef void copy_file_block2buffer(self, const char* filename, int32_t sV0, int32_t sV1, int32_t sH0, int32_t sH1, int32_t sD0, int32_t sD1,
                                     uint8_t * buf, uint32_t pxl_size, int64_t offs, int64_t stridex,
                                     int64_t stridexy, int64_t stridexyz):
        raise NotImplementedError

    cdef void copy_block2sub_buf(self, uint8_t* src, uint8_t* dst,
                           int32_t dimi, int32_t dimj, int32_t dimk, int32_t typesize,
                           int64_t s_stridej, int64_t s_strideij,
                           int64_t d_stridej, int64_t d_strideij):

        cdef int32_t i, k, j
        cdef int64_t s_index, d_index
        cdef uint8_t* s_slice = src
        cdef uint8_t* d_slice = dst
        cdef uint8_t * s_stripe
        cdef uint8_t * d_stripe

        s_stridej *= typesize
        s_strideij *= typesize
        d_stridej *= typesize
        d_strideij *= typesize

        for k in range(dimk):
            s_stripe = s_slice
            d_stripe  = d_slice
            for i in range(dimi):
                memcpy(d_stripe, s_stripe, dimj * typesize)
                s_stripe += s_stridej
                d_stripe += d_stridej
            s_slice += s_strideij
            d_slice += d_strideij

    cdef void copy_rgb_block2vaa3d_raw_sub_buf(self, uint8_t* src, uint8_t* dst,
                                                 int32_t dimi, int32_t dimj, int32_t dimk, int32_t typesize,
                                                 int64_t s_stridej, int64_t s_strideij,
                                                 int64_t d_stridej, int64_t d_strideij, int64_t d_strideijk):

        cdef:
            int32_t c, i, j, k
            int64_t s_index, d_index
            uint8_t * s_slice
            uint8_t * d_slice
            uint8_t * s_stripe
            uint8_t * d_stripe
            uint16_t * src16
            uint16_t * dst16
            uint16_t * s_slice16
            uint16_t * d_slice16
            uint16_t * s_stripe16
            uint16_t * d_stripe16

        if typesize == 1:
            for c in range(3):
                s_slice = src
                d_slice = dst
                for k in range(dimk):
                    s_stripe = s_slice
                    d_stripe = d_slice
                    for i in range(dimi):
                        for j in range(dimj):
                            d_stripe[j] = s_stripe[j * 3]
                        s_stripe += s_stridej * 3
                        d_stripe += d_stridej
                    s_slice += s_strideij * 3
                    d_slice += d_strideij
                src += 1
                dst += d_strideijk
        else:  # must be guaranteed that typesize == 2
            for c in range(3):
                s_slice16 = <uint16_t*>src
                d_slice16 = <uint16_t*>dst
                for k in range(dimk):
                    s_stripe16 = s_slice16
                    d_stripe16 = d_slice16
                    for i in range(dimi):
                        for j in range(dimj):
                            d_stripe16[j] = s_stripe16[j * 3]
                        s_stripe16 += s_stridej * 3
                        d_stripe16 += d_stridej
                    s_slice16 += s_strideij * 3
                    d_slice16 += d_strideij
                src16 += 1
                dst16 += d_strideijk



cdef class Tiff3DFmtMngr(VirtualFmtMngr):

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void copy_file_block2buffer(self, const char* filename,
                                     int32_t sV0, int32_t sV1, int32_t sH0, int32_t sH1, int32_t sD0, int32_t sD1,
                                     uint8_t* buf, uint32_t pxl_size,
                                     int64_t offs,
                                     int64_t stridex,
                                     int64_t stridexy,
                                     int64_t stridexyz):
        cdef:
            unsigned int sz[4]
            int datatype = 0
            bint b_swap = 0
            void* fhandle = load_tiff3d2metadata(filename, sz[0], sz[1], sz[2], sz[3], datatype, b_swap)

        if datatype != pxl_size:
            raise IOError("Tiff3DFmtMngr.copy_file_block2buffer: source data type differs from destination pixel size.")

        cdef cnp.ndarray[cnp.uint8_t, ndim=1] npbuf_t = np.zeros(sz[0] * sz[1] * (sD1 - sD0) * sz[3] * datatype, dtype=np.uint8)
        cdef unsigned char * buf_t = <uint8_t *> npbuf_t.data

        read_tiff_3d_file_to_buffer(fhandle, buf_t, sz[0], sz[1], sD0, sD1 - 1, b_swap,1, -1, -1, -1, -1)
        close_tiff3d_file(fhandle)

        cdef:
            int64_t s_stridej = sz[0]
            int64_t s_strideij = sz[0] * sz[1]
            int64_t d_stridej = stridex
            int64_t d_strideij = stridexy
            int64_t d_strideijk = stridexyz
            int32_t dimi = sV1 - sV0
            int32_t dimj = sH1 - sH0
            int32_t dimk = sD1 - sD0

        if sz[3] == 1:  # single channel Tiff
            self.copy_block2sub_buf(
                buf_t + pxl_size * (s_stridej * sV0 + sH0),
                buf + pxl_size * offs,
                dimi, dimj, dimk, pxl_size,
                s_stridej, s_strideij, d_stridej, d_strideij
            )
        elif sz[3] == 3:  # RGB Tiff
            self.copy_rgb_block2vaa3d_raw_sub_buf(
                buf_t + 3 * pxl_size * (s_stridej * sV0 + sH0),
                buf + pxl_size * offs,
                dimi, dimj, dimk, pxl_size,
                s_stridej, s_strideij,
                d_stridej, d_strideij, d_strideijk)
        else:
            raise IOError("Tiff3DFmtMngr.copy_file_block2buffer: unsupported number of channels.")


from libc.stdint cimport uint8_t, int64_t, int32_t, uint32_t


cdef class VirtualFmtMngr:
    cdef public void copy_file_block2buffer(self, const char* filename, int32_t sV0, int32_t sV1, int32_t sH0, int32_t sH1, int32_t sD0, int32_t sD1,
                                             uint8_t * buf, uint32_t pxl_size, int64_t offs, int64_t stridex,
                                             int64_t stridexy, int64_t stridexyz)

    cdef void copy_block2sub_buf(self, uint8_t * src, uint8_t * dst,
                                int32_t dimi, int32_t dimj, int32_t dimk, int32_t typesize,
                                int64_t s_stridej, int64_t s_strideij,
                                int64_t d_stridej, int64_t d_strideij)

    cdef void copy_rgb_block2vaa3d_raw_sub_buf(self, uint8_t * src, uint8_t * dst,
                                              int32_t dimi, int32_t dimj, int32_t dimk, int32_t typesize,
                                              int64_t s_stridej, int64_t s_strideij,
                                              int64_t d_stridej, int64_t d_strideij, int64_t d_strideijk)


cdef class Tiff3DFmtMngr(VirtualFmtMngr):
    cdef public void copy_file_block2buffer(self, const char* filename, int32_t sV0, int32_t sV1, int32_t sH0, int32_t sH1, int32_t sD0, int32_t sD1,
                                            uint8_t* buf, uint32_t pxl_size, int64_t offs, int64_t stridex,
                                            int64_t stridexy, int64_t stridexyz)


"""*================================================================
*   Copyright (C) 2021 Yufeng Liu (Braintell, Southeast University). All rights reserved.
*   
*   Filename    : swc_handler.py
*   Author      : Yufeng Liu
*   Date        : 2021-03-15
*   Description : 
*
================================================================*"""
import re
import numpy as np
from copy import deepcopy

NEURITE_TYPES = {
    'soma': [1],
    'axon': [2],
    'basal dendrite': [3],
    'apical dendrite': [4],
    'dendrite': [3,4],
}


def load_spacings(spacing_file, zxy_order=False):
    """
    Load the spacing information for each brain. The spacing here refers to
    the resolution along x,y,z axes.
    """
    spacing_dict = {}
    with open(spacing_file, 'r') as fp:
        for line in fp.readlines():
            line = line.strip()
            if not line: continue
            ctxts = line.split(',')
            brain_id = ctxts[0]
            if not brain_id.isdigit(): 
                continue # the brain is encoded as digits

            brain_id = int(brain_id)
            spacing = tuple(map(float, ctxts[1:]))
            if zxy_order:
                spacing = (spacing[2],spacing[0],spacing[1])
            spacing_dict[brain_id] = spacing
    
    return spacing_dict


def parse_swc(swc_file):
    tree = []
    with open(swc_file) as fp:
        for line in fp.readlines():
            line = line.strip()
            if not line: continue
            if line[0] == '#': continue
            idx, type_, x, y, z, r, p = line.split()[:7]
            idx = int(idx)
            type_ = int(type_)
            x = float(x)
            y = float(y)
            z = float(z)
            r = float(r)
            p = int(p)
            tree.append((idx, type_, x, y, z, r, p))
    
    return tree


def write_swc(tree, swc_file, header=tuple()):
    if header is None:
        header = []
    with open(swc_file, 'w') as fp:
        for s in header:
            if not s.startswith("#"):
                s = "#" + s
            if not s.endswith("\n") or not s.endswith("\r"):
                s += "\n"
            fp.write(s)
        fp.write(f'##n type x y z r parent\n')
        for leaf in tree:
            idx, type_, x, y, z, r, p = leaf
            fp.write(f'{idx:d} {type_:d} {x:.5f} {y:.5f} {z:.5f} {r:.1f} {p:d}\n')


def find_soma_node(tree, p_soma=-1, p_idx_in_leaf=6):
    for leaf in tree:
        if leaf[p_idx_in_leaf] == p_soma:
            #print('Soma: ', leaf)
            return leaf[0]
    #raise ValueError("Could not find the soma node!")
    return -99


def find_soma_index(tree, p_soma=-1):
    for i, leaf in enumerate(tree):
        if leaf[6] == p_soma:
            return i
    #raise ValueError("find_soma_index: Could not find the somma node!")
    return -99


def get_child_dict(tree, p_idx_in_leaf=6):
    child_dict = {}
    for leaf in tree:
        p_idx = leaf[p_idx_in_leaf]
        if p_idx in child_dict:
            child_dict[p_idx].append(leaf[0])
        else:
            child_dict[p_idx] = [leaf[0]]
    return child_dict


def get_index_dict(tree):
    index_dict = {}
    for i, leaf in enumerate(tree):
        idx = leaf[0]
        index_dict[idx] = i
    return index_dict


def is_in_box(x, y, z, imgshape):
    """
    imgshape must be in (z,y,x) order
    """
    if x < 0 or y < 0 or z < 0 or \
        x > imgshape[2] - 1 or \
        y > imgshape[1] - 1 or \
        z > imgshape[0] - 1:
        return False
    return True

def is_in_bbox(x, y, z, zyxzyx):
    """ 
    zyxzyx is bbox in format of [(zmin, ymin, xmin), (zmax, ymax, xmax)]
    """
    (zmin, ymin, xmin), (zmax, ymax, xmax) = zyxzyx
    if x < xmin or y < ymin or z < zmin or \
        x > xmax or \
        y > ymax or \
        z > zmax:
        return False
    return True

def prune(tree: list, ind_set: set):
    """
    prune all nodes given by ind_set in morph
    """
    child_dict = get_child_dict(tree)
    index_dict = get_index_dict(tree)
    tree = deepcopy(tree)
    for i in ind_set:
        q = []
        ind = index_dict[i]
        if tree[ind] is None:
            continue
        tree[ind] = None
        if i in child_dict:
            q.extend(child_dict[i])
        while len(q) > 0:
            head = q.pop(0)
            ind = index_dict[head]
            if tree[ind] is None:
                continue
            tree[ind] = None
            if head in child_dict:
                q.extend(child_dict[head])
    return [t for t in tree if t is not None]


def trim_swc(tree_orig, imgshape, keep_candidate_points=True, bfs=True):
    """
    Trim the out-of-box and non_connecting leaves
    """
    if bfs:
        ib = set(t[0] for t in tree_orig if is_in_box(*t[2:5], imgshape))
        if keep_candidate_points:
            child_dict = get_child_dict(tree_orig)
            ib = ib.union(*(child_dict[i] for i in ib if i in child_dict))
        return prune(tree_orig, set(t[0] for t in tree_orig) - ib)

    def traverse_leaves(idx, child_dict, good_points, cand_pints, pos_dict):
        leaf = pos_dict[idx]
        p_idx, ib = leaf[-2:]

        if (p_idx in good_points) or (p_idx == -1):
            if ib:
                good_points.add(idx)    # current node
            else:
                cand_points.add(idx)
                return

        if idx not in child_dict:
            return

        for new_idx in child_dict[idx]:
            traverse_leaves(new_idx, child_dict, good_points, cand_pints, pos_dict)

    # execute trimming
    pos_dict = {}
    tree = deepcopy(tree_orig)
    for i, leaf in enumerate(tree_orig):
        idx, type_, x, y, z, r, p = leaf
        leaf = (idx, type_, x, y, z, r, p, is_in_box(x,y,z,imgshape))
        pos_dict[idx] = leaf
        tree[i] = leaf

    good_points = set() # points and all its upstream parents are in-box
    cand_points = set() # all upstream parents are in-box, itself not
    # initialize the visited set with soma, whose parent index is -1
    soma_idx = None
    for leaf in tree:
        if leaf[-2] == -1:
            soma_idx = leaf[0]
            break
    #print(soma_idx)

    child_dict = {}
    for leaf in tree:
        if leaf[-2] in child_dict:
            child_dict[leaf[-2]].append(leaf[0])
        else:
            child_dict[leaf[-2]] = [leaf[0]]
    # do DFS searching
    #print(soma_idx)
    traverse_leaves(soma_idx, child_dict, good_points, cand_points, pos_dict)
    #print("#good/#cand/#total:", len(good_points), len(cand_points), len(pos_dict))  

    # return the tree, (NOTE: without order)
    tree_trim = []
    if keep_candidate_points:
        keep_points = good_points | cand_points
    else:
        keep_points = good_points

    for i, leaf in enumerate(tree):
        idx = leaf[0]
        if idx in keep_points:
            tree_trim.append(leaf[:-1])

    return tree_trim


def trim_out_of_box(tree_orig, imgshape, keep_candidate_points=True):
    """ 
    Trim the out-of-box leaves
    """
    # execute trimming
    child_dict = {}
    for leaf in tree_orig:
        if leaf[-1] in child_dict:
            child_dict[leaf[-1]].append(leaf[0])
        else:
            child_dict[leaf[-1]] = [leaf[0]]
    
    pos_dict = {}
    for i, leaf in enumerate(tree_orig):
        pos_dict[leaf[0]] = leaf

    tree = []
    for i, leaf in enumerate(tree_orig):
        idx, type_, x, y, z, r, p = leaf
        ib = is_in_box(x,y,z,imgshape)
        if ib: 
            tree.append(leaf)
        elif keep_candidate_points:
            if p in pos_dict and is_in_box(*pos_dict[p][2:5], imgshape):
                tree.append(leaf)
            elif idx in child_dict:
                for ch_leaf in child_dict[idx]:
                    if is_in_box(*pos_dict[ch_leaf][2:5], imgshape):
                        tree.append(leaf)
                        break
    return tree


def get_specific_neurite(tree, type_id):
    if (not isinstance(type_id, list)) and (not isinstance(type_id, tuple)):
        type_id = (type_id,)
    
    new_tree = []
    for leaf in tree:
        if leaf[1] in type_id:
            new_tree.append(leaf)
    return new_tree


def shift_swc(swc_file, sx, sy, sz):
    if type(swc_file) == list:
        tree = swc_file
    else:
        tree = parse_swc(swc_file)
    new_tree = []
    for node in tree:
        idx, type_, x, y, z, r, p = node
        x = x - sx
        y = y - sy
        z = z - sz
        node = (idx, type_, x, y, z, r, p)
        new_tree.append(node)
    return new_tree


def scale_swc(swc_file, scale):
    if type(swc_file) == list:
        tree = swc_file
    else:
        tree = parse_swc(swc_file)
    if isinstance(scale, (int, float)):
        scale_x, scale_y, scale_z = scale, scale, scale
    elif isinstance(scale, tuple) or isinstance(scale, list):
        scale_x, scale_y, scale_z = scale
    else:
        raise NotImplementedError(f"Type of parameter scale {type(scale)} is not supported!")

    new_tree = []
    for node in tree:
        idx, type_, x, y, z, r, p = node
        x *= scale_x
        y *= scale_y
        z *= scale_z
        node = (idx, type_, x, y, z, r, p)
        new_tree.append(node)
    return new_tree

def flip_swc(swc_file, axis='y', dim=None):
    if type(swc_file) == list:
        tree = swc_file
    else:
        tree = parse_swc(swc_file)
    
    new_tree = []
    for node in tree:
        idx, type_, x, y, z, r, p = node
        if axis == 'x':
            x = dim - x
        elif axis == 'y':
            y = dim - y
        elif axis == 'z':
            z = dim - z
        node = (idx, type_, x, y, z, r, p)
        new_tree.append(node)
    return new_tree

def crop_tree_by_bbox(morph, bbox, keep_candidate_points=True):
    """ 
    Crop swc by trim all nodes out-of-bbox. This function differs from `trim_out_of_box` it does
    not assume center cropping
    """
    if isinstance(morph, list):
        mtree = morph
    else:
        mtree = morph.tree

    tree = []
    for i, leaf in enumerate(morph.tree):
        idx, type_, x, y, z, r, p = leaf[:7]
        ib = is_in_bbox(x,y,z,bbox) 
        if ib: 
            tree.append(leaf)
            if keep_candidate_points and (idx in morph.child_dict):
                for ch_leaf in morph.child_dict[idx]:
                    if not is_in_bbox(*morph.pos_dict[ch_leaf][2:5], bbox):
                        tree.append(morph.pos_dict[ch_leaf])
    return tree


def rm_disconnected(tree: list, anchor: int):
    roots = [t[0] for t in tree if t[6] == -1]
    ch = get_child_dict(tree)
    idx = get_index_dict(tree)
    flag = np.zeros(len(tree), dtype=int)
    for r in roots:
        q = [r]
        while len(q) > 0:
            head = q.pop(0)
            flag[idx[head]] = r
            if head in ch:
                q.extend(ch[head])
    ind = flag[idx[anchor]]
    return prune(tree, set(t[0] for t, f in zip(tree, flag) if f != ind))

def get_soma_from_swc(swcfile):
    # fast parse swc information
    # only for swc, not eswc
    with open(swcfile) as fp:
        soma_str = re.search('.* -1\n', fp.read()).group()
    soma = soma_str.split()
    return soma




from setuptools import Extension, setup
from pathlib import Path
import numpy as np
import cmake_build_extension
import importlib
import platform
import subprocess
import os
import shutil
from importlib.util import find_spec
from setuptools.command.build_ext import build_ext


class MyBuildExtension(build_ext):
    def initialize_options(self):

        # Initialize base class
        build_ext.initialize_options(self)

        # Initialize the '--define' custom option, overriding the pre-existing one.
        # Originally, it was aimed to pass C preprocessor definitions, but instead we
        # use it to pass custom configuration options to CMake.
        self.define = None

        # Initialize the '--component' custom option.
        # It overrides the content of the cmake_component option of CMakeExtension.
        self.component = None

        # Initialize the 'no-cmake-extension' custom option.
        # It allows disabling one or more CMakeExtension from the command line.
        self.no_cmake_extension = None

    @staticmethod
    def extend_cmake_prefix_path(path: str) -> None:

        abs_path = Path(path).absolute()

        if not abs_path.exists():
            raise ValueError(f"Path {abs_path} does not exist")

        if "CMAKE_PREFIX_PATH" in os.environ:
            os.environ[
                "CMAKE_PREFIX_PATH"
            ] = f"{str(path)}:{os.environ['CMAKE_PREFIX_PATH']}"
        else:
            os.environ["CMAKE_PREFIX_PATH"] = str(path)

    def finalize_options(self):

        # Parse the custom CMake options and store them in a new attribute
        defines = [] if self.define is None else self.define.split(";")
        self.cmake_defines = [f"-D{define}" for define in defines]

        # Parse the disabled CMakeExtension modules and store them in a new attribute
        self.no_cmake_extensions = (
            []
            if self.no_cmake_extension is None
            else self.no_cmake_extension.split(";")
        )

        # Call base class
        build_ext.finalize_options(self)

    def run(self) -> None:
        """
        Process all the registered extensions executing only the CMakeExtension objects.
        """

        # Filter the CMakeExtension objects
        cmake_extensions = [e for e in self.extensions if isinstance(e, cmake_build_extension.CMakeExtension)]

        if len(cmake_extensions) == 0:
            raise ValueError("No CMakeExtension objects found")

        # Check that CMake is installed
        if shutil.which("cmake") is None:
            raise RuntimeError("Required command 'cmake' not found")

        # Check that Ninja is installed
        if shutil.which("ninja") is None:
            raise RuntimeError("Required command 'ninja' not found")

        for ext in cmake_extensions:

            # Disable the extension if specified in the command line
            if (
                ext.name in self.no_cmake_extensions
                or "all" in self.no_cmake_extensions
            ):
                continue

            # Disable all extensions if this env variable is present
            disabled_set = {"0", "false", "off", "no"}
            env_var_name = "CMAKE_BUILD_EXTENSION_ENABLED"
            if (
                env_var_name in os.environ
                and os.environ[env_var_name].lower() in disabled_set
            ):
                continue

            self.my_build_extension(ext)
        self.extensions = [e for e in self.extensions if not isinstance(e, cmake_build_extension.CMakeExtension)]
        build_ext.run(self)

    # don't use Ninja
    def my_build_extension(self, ext: cmake_build_extension.CMakeExtension) -> None:
        """
        Build a CMakeExtension object.

        Args:
            ext: The CMakeExtension object to build.
        """

        # if self.inplace and ext.disable_editable:
        #     print(f"Editable install recognized. Extension '{ext.name}' disabled.")
        #     return

        # Export CMAKE_PREFIX_PATH of all the dependencies
        for pkg in ext.cmake_depends_on:

            try:
                importlib.import_module(pkg)
            except ImportError:
                raise ValueError(f"Failed to import '{pkg}'")

            init = find_spec(pkg).origin
            cmake_build_extension.BuildExtension.extend_cmake_prefix_path(path=str(Path(init).parent))

        cmake_install_prefix = ext.install_prefix

        # CMake configure arguments
        configure_args = [
            # '-GNinja',
            f"-DCMAKE_BUILD_TYPE={ext.cmake_build_type}",
            f"-DCMAKE_INSTALL_PREFIX:PATH={cmake_install_prefix}",
            # Fix #26: https://github.com/diegoferigo/cmake-build-extension/issues/26
            # f"-DCMAKE_MAKE_PROGRAM={shutil.which('ninja')}",
        ]

        # Extend the configure arguments with those passed from the extension
        configure_args += ext.cmake_configure_options

        # CMake build arguments
        build_args = ["--config", ext.cmake_build_type]

        if platform.system() == "Windows":

            configure_args += []

        elif platform.system() in {"Linux", "Darwin"}:

            configure_args += ['-DCMAKE_C_FLAGS="-fPIC"']

        else:
            raise RuntimeError(f"Unsupported '{platform.system()}' platform")

        # Parse the optional CMake options. They can be passed as:
        #
        # python setup.py build_ext -D"BAR=Foo;VAR=TRUE"
        # python setup.py bdist_wheel build_ext -D"BAR=Foo;VAR=TRUE"
        # python setup.py install build_ext -D"BAR=Foo;VAR=TRUE"
        # python setup.py install -e build_ext -D"BAR=Foo;VAR=TRUE"
        # pip install --global-option="build_ext" --global-option="-DBAR=Foo;VAR=TRUE" .
        #
        configure_args += self.cmake_defines

        # Get the absolute path to the build folder
        build_folder = str(Path(".").absolute() / f"{self.build_temp}_{ext.name}")

        # Make sure that the build folder exists
        Path(build_folder).mkdir(exist_ok=True, parents=True)

        # 1. Compose CMake configure command
        configure_command = [
                                "cmake",
                                "-S",
                                ext.source_dir,
                                "-B",
                                build_folder,
                            ] + configure_args

        # 2. Compose CMake build command
        build_command = ["cmake", "--build", build_folder] + build_args

        # 3. Compose CMake install command
        install_command = ["cmake", "--install", build_folder]

        # If the cmake_component option of the CMakeExtension is used, install just
        # the specified component.
        if self.component is None and ext.cmake_component is not None:
            install_command.extend(["--component", ext.cmake_component])

        # Instead, if the `--component` command line option is used, install just
        # the specified component. This has higher priority than what specified in
        # the CMakeExtension.
        if self.component is not None:
            install_command.extend(["--component", self.component])

        print("")
        print("==> Configuring:")
        print(f"$ {' '.join(configure_command)}")
        print("")
        print("==> Building:")
        print(f"$ {' '.join(build_command)}")
        print("")
        print("==> Installing:")
        print(f"$ {' '.join(install_command)}")
        print("")

        # Call CMake
        subprocess.check_call(configure_command)
        subprocess.check_call(build_command)
        subprocess.check_call(install_command)

        # Write content to the top-level __init__.py
        if ext.write_top_level_init is not None:
            with open(file=Path(cmake_install_prefix) / "__init__.py", mode="w") as f:
                f.write(ext.write_top_level_init)

        # Write content to the bin/__main__.py magic file to expose binaries
        if len(ext.expose_binaries) > 0:
            bin_dirs = {str(Path(d).parents[0]) for d in ext.expose_binaries}

            import inspect

            main_py = inspect.cleandoc(
                f"""
                from pathlib import Path
                import subprocess
                import sys

                def main():

                    binary_name = Path(sys.argv[0]).name
                    prefix = Path(__file__).parent.parent
                    bin_dirs = {str(bin_dirs)}

                    binary_path = ""

                    for dir in bin_dirs:
                        path = prefix / Path(dir) / binary_name
                        if path.is_file():
                            binary_path = str(path)
                            break

                        path = Path(str(path) + ".exe")
                        if path.is_file():
                            binary_path = str(path)
                            break

                    if not Path(binary_path).is_file():
                        name = binary_path if binary_path != "" else binary_name
                        raise RuntimeError(f"Failed to find binary: {{ name }}")

                    sys.argv[0] = binary_path

                    result = subprocess.run(args=sys.argv, capture_output=False)
                    exit(result.returncode)

                if __name__ == "__main__" and len(sys.argv) > 1:
                    sys.argv = sys.argv[1:]
                    main()"""
            )

            bin_folder = cmake_install_prefix / "bin"
            Path(bin_folder).mkdir(exist_ok=True, parents=True)
            with open(file=bin_folder / "__main__.py", mode="w") as f:
                f.write(main_py)


extensions = [
    Extension(
        'v3dpy.neuron_utilities.profiling',
        ['v3dpy/neuron_utilities/profiling.pyx'],
        include_dirs=[np.get_include()],
        language='c++'
    ),
    Extension(
        'v3dpy.loaders.pbd',
        ['v3dpy/loaders/pbd.pyx'],
        include_dirs=[np.get_include()],
        language='c++'
    ),
    Extension(
        'v3dpy.loaders.raw',
        ['v3dpy/loaders/raw.pyx'],
        include_dirs=[np.get_include()],
        language='c++'
    ),
    Extension(
        'v3dpy.terafly.tiff_manage',
        ['v3dpy/terafly/tiff_manage.pyx'],
        include_dirs=['3rdparty/libtiff/include', np.get_include()],
        library_dirs=['3rdparty/libtiff/lib', '3rdparty/libtiff/lib64'],
        libraries=['tiff'],
        language='c++'
    ),
    Extension(
        'v3dpy.terafly.format_managers',
        ['v3dpy/terafly/format_managers.pyx'],
        include_dirs=[np.get_include()],
        language='c++'
    ),
    Extension(
        'v3dpy.terafly.volume_managers',
        ['v3dpy/terafly/volume_managers.pyx'],
        include_dirs=[np.get_include()],
        language='c++',
    ),
]

setup(
    ext_modules=[
        cmake_build_extension.CMakeExtension(
        'tiff',
        '3rdparty/libtiff',
        source_dir=str(Path('3rdparty/libtiff').absolute()),
        cmake_configure_options=['-DBUILD_SHARED_LIBS=OFF', '-Dtiff-tests=OFF', '-Dtiff-tools=OFF', '-Dtiff-docs=OFF', '-Dmdi=OFF',
                                '-Djpeg=OFF', '-Dzlib=OFF', '-Dzstd=OFF', '-Dlzma=OFF', '-Djbig=OFF', '-Dwebp=OFF']
    )] + extensions,
    cmdclass=dict(
        # Enable the CMakeExtension entries defined above
        build_ext=MyBuildExtension,
        # If the setup.py or setup.cfg are in a subfolder wrt the main CMakeLists.txt,
        # you can use the following custom command to create the source distribution.
        # sdist=cmake_build_extension.GitSdistFolder
    ),
)



"""
Loaders for Vaa3D image formats. To use them, intialize an IO object which can be fed some loading and saving options.
Their member functions can be called to load and save images.
"""

from .raw import Raw
from .pbd import PBD

__all__ = ['Raw', 'PBD']


cimport cython
from pathlib import Path
from .config import *
import struct
import numpy as np
cimport numpy as cnp

from libcpp.string cimport string
from libcpp.vector cimport vector
from .format_managers cimport Tiff3DFmtMngr, VirtualFmtMngr
from libc.stdint cimport uint32_t, int64_t, int32_t, uint16_t


cdef class VirtualVolume:
    cdef:
        public object root_dir
        public float VXL_V, VXL_H, VXL_D
        public float ORG_V, ORG_H, ORG_D
        public uint32_t DIM_V, DIM_H, DIM_D, DIM_C
        public uint32_t BYTESxCHAN
    def __cinit__(self):
        self.ORG_V = self.ORG_H = self.ORG_D = 0.0
        self.DIM_V = self.DIM_H = self.DIM_D = self.DIM_C = 0
        self.BYTESxCHAN = 0

    def __init__(self, root_dir: Path = None, vxl_1=0, vxl_2=0, vxl_3=0):
        self.root_dir = root_dir
        self.VXL_V = vxl_1
        self.VXL_H = vxl_2
        self.VXL_D = vxl_3


    cpdef cnp.ndarray load_sub_volume(self, int32_t v0=-1, int32_t v1=-1, int32_t h0=-1,
                                    int32_t h1=-1, int32_t d0=-1, int32_t d1=-1):
        raise NotImplementedError



cdef struct Rect:
    int32_t H0
    int32_t H1
    int32_t V0
    int32_t V1


cdef struct Segm:
    uint32_t D0
    uint32_t D1
    uint32_t ind0
    uint32_t ind1


cdef class Block(VirtualVolume):
    cdef:
        public str DIR_NAME
        public object CONTAINER
        public uint16_t ROW_INDEX, COL_INDEX
        public uint32_t HEIGHT, WIDTH, DEPTH, N_BLOCKS, N_CHANS, N_BYTESxCHAN
        public int32_t ABS_V, ABS_H
        public vector[string] FILENAMES
        public vector[uint32_t] BLOCK_SIZE
        public vector[int32_t] BLOCK_ABS_D

    def __cinit__(self):
        self.BLOCK_SIZE = vector[uint32_t]()
        self.BLOCK_ABS_D = vector[int32_t]()
        self.FILENAMES = vector[string]()
        self.HEIGHT = self.WIDTH = self.DEPTH = self.N_BLOCKS = self.N_CHANS \
            = self.ABS_V = self.ABS_H = 0
        self.DIR_NAME= ''
        self.N_BYTESxCHAN = 0

    def __init__(self, object container, uint16_t row_index, uint16_t col_index, object file):
        super(Block, self).__init__()
        self.CONTAINER = container
        self.ROW_INDEX = row_index
        self.COL_INDEX = col_index
        self.unbinarize_from(file)

    cdef void unbinarize_from(self, object file):
        cdef uint32_t i
        cdef uint16_t str_size
        self.HEIGHT, self.WIDTH, self.DEPTH, self.N_BLOCKS, self.N_CHANS, self.ABS_V, self.ABS_H, str_size = \
            struct.unpack('IIIIIiiH', file.read(30))
        self.DIR_NAME = file.read(str_size).decode('utf-8').rstrip("\x00")
        for i in range(self.N_BLOCKS):
            str_size = struct.unpack('H', file.read(2))[0]
            self.FILENAMES.push_back(file.read(str_size).rstrip(b"\x00"))
            self.BLOCK_SIZE.push_back(struct.unpack('I', file.read(4))[0])
            self.BLOCK_ABS_D.push_back(struct.unpack('i', file.read(4))[0])
        self.N_BYTESxCHAN = struct.unpack('I', file.read(4))[0]

    @cython.boundscheck(False)
    @cython.nonecheck(False)
    cpdef bytes get_fmt(self):
        cdef bytes temp = bytes(self.FILENAMES[0]).split(b'.')[-1]
        if temp == b'tif' or temp == b'tif':
            return b'Tiff3D'
        elif temp == b'v3draw':
            return b'Vaa3DRaw'
        else:
            raise IOError(f'Block: Unknown file format {temp}')


cdef bint intersects_segm(Block block, int32_t d0, int32_t d1, Segm& out):
    cdef int32_t i0 = 0, i1 = block.N_BLOCKS - 1

    if d0 >= <int32_t>(block.BLOCK_ABS_D[i1] + block.BLOCK_SIZE[i1]) or d1 <= 0:
        return False # no intersection

    while i0 < i1:
        if d0 < block.BLOCK_ABS_D[i0 + 1]:
            break
        i0 += 1

    while i1 > 0:
        if d1 > block.BLOCK_ABS_D[i1]:
            break
        i1 -= 1

    out.D0 = max(d0, 0)
    out.D1 = min(d1, <int32_t>block.DEPTH)
    out.ind0 = i0
    out.ind1 = i1

    return True

cdef bint intersects_rect(Block block, const Rect& area, Rect& out):
    cdef:
        int32_t t1 = block.ABS_H + block.WIDTH
        int32_t t2 = block.ABS_V + block.HEIGHT

    if area.H0 < t1 and area.H1 > block.ABS_H and area.V0 < t2 and area.V1 > block.ABS_V:

        out.H0 = max(block.ABS_H, area.H0)
        out.V0 = max(block.ABS_V, area.V0)
        out.H1 = min(t1, area.H1)
        out.V1 = min(t2, area.V1)

        return True
    else:
        return False


cdef class TiledVolume(VirtualVolume):
    cdef:
        uint16_t N_ROWS, N_COLS  # <-- Static type declarations
        list BLOCKS  # <-- Declare a 2D C-style array of Blocks
        int32_t reference_system_first, reference_system_second, reference_system_thrid
        float VXL_1, VXL_2, VXL_3

    def __cinit__(self):
        self.reference_system_first = self.reference_system_second = self.reference_system_thrid = \
            self.VXL_1 = self.VXL_2 = self.VXL_3 = self.N_ROWS = self.N_COLS = 0

    def __init__(self, root_dir: Path):
        super(TiledVolume, self).__init__(root_dir)
        self.BLOCKS = None
        mdata_filepath = root_dir / MDATA_BIN_FILE_NAME
        if mdata_filepath.is_file():  # We need to convert string back to Path object for is_file()
            self.load(mdata_filepath)
            self.init_channels()
        else:
            raise ValueError(f"TiledVolume: unable to find metadata file at {mdata_filepath}")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef load(self, mdata_filepath: Path):
        cdef uint16_t i, j

        with open(mdata_filepath, 'rb') as f:
            mdata_version_read = struct.unpack('f', f.read(4))[0]
            if mdata_version_read != MDATA_BIN_FILE_VERSION:
                f.seek(0)
                str_size = struct.unpack('H', f.read(2))[0]
                f.read(str_size)
            self.reference_system_first, self.reference_system_second, self.reference_system_thrid, \
                self.VXL_1, self.VXL_2, self.VXL_3, self.VXL_V, self.VXL_H, self.VXL_D, \
                self.ORG_V, self.ORG_H, self.ORG_D, self.DIM_V, self.DIM_H, self.DIM_D, \
                self.N_ROWS, self.N_COLS = struct.unpack('iiifffffffffIIIHH', f.read(64))

            self.BLOCKS = [[Block(self, i, j, f) for j in range(self.N_COLS)] for i in range(self.N_ROWS)]

    cdef init_channels(self):
        self.DIM_C = self.BLOCKS[0][0].N_CHANS
        self.BYTESxCHAN = self.BLOCKS[0][0].N_BYTESxCHAN

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cpdef cnp.ndarray load_sub_volume(self, int32_t V0=-1, int32_t V1=-1, int32_t H0=-1,
                                      int32_t H1=-1, int32_t D0=-1, int32_t d1=-1):
        cdef:
            uint16_t row, col
            int32_t k, sV0, sV1, sH0, sH1, sD0, sD1, bV0, bH0, bD0
            Block block
            Rect intersect_area
            Segm intersect_segm
            bytes slice_fullpath
            bint first_time = True
            bytes ffmt

        V0, H0, D0 = max(0, V0), max(0, H0), max(0, D0)
        V1 = V1 if 0 <= V1 <= <int32_t>self.DIM_V else self.DIM_V
        H1 = H1 if 0 <= H1 <= <int32_t>self.DIM_H else self.DIM_H
        d1 = d1 if 0 <= d1 <= <int32_t>self.DIM_D else self.DIM_D
        assert V1 > V0 and H1 > H0 and d1 > D0, "TiledVolume: The start position should be lower than the end position."

        cdef:
            int64_t sbv_height = V1 - V0
            int64_t sbv_width = H1 - H0
            int64_t sbv_depth = d1 - D0
            int64_t sbv_channels
            int64_t sbv_bytes_chan
            cnp.ndarray npsubvol
            unsigned char* subvol
            VirtualFmtMngr fmt_mngr

        subvol_area = Rect()
        subvol_area.H0 = H0
        subvol_area.H1 = H1
        subvol_area.V0 = V0
        subvol_area.V1 = V1

        ffmt = self.BLOCKS[0][0].get_fmt()
        if ffmt == b'Tiff3D':
            fmt_mngr = Tiff3DFmtMngr()
        elif ffmt == b'Vaa3DRaw':
            raise NotImplementedError

        if intersects_segm(self.BLOCKS[0][0], D0, d1, intersect_segm):
            for row in range(self.N_ROWS):
                for col in range(self.N_COLS):
                    block = self.BLOCKS[row][col]
                    if intersects_rect(block, subvol_area, intersect_area):
                        for k in range(intersect_segm.ind0, intersect_segm.ind1 + 1):
                            if first_time:
                                first_time = False
                                sbv_channels = self.DIM_C
                                sbv_bytes_chan = self.BYTESxCHAN
                                assert sbv_channels == 1 , "TiledVolume: Multi channel not supported yet."
                                if sbv_bytes_chan == 1:
                                    dt = np.uint8
                                elif sbv_bytes_chan == 2:
                                    dt = np.uint16
                                elif sbv_bytes_chan == 4:
                                    dt = np.float32
                                else:
                                    raise ValueError(f"TiledVolume: Unsupported Datatype {self.BYTESxCHAN}")
                                npsubvol = np.zeros((sbv_channels, sbv_depth, sbv_height, sbv_width), dtype=dt)
                                subvol = <unsigned char *> npsubvol.data

                            slice_fullpath = str(self.root_dir / block.DIR_NAME /
                                                 bytes(block.FILENAMES[k]).decode('utf-8')).encode('utf-8')

                            # vertices of file block
                            sV0 = 0 if V0 < intersect_area.V0 else V0 - block.ABS_V
                            sV1 = block.HEIGHT if V1 > intersect_area.V1 else V1 - block.ABS_V
                            sH0 = 0 if H0 < intersect_area.H0 else H0 - block.ABS_H
                            sH1 = block.WIDTH if H1 > intersect_area.H1 else H1 - block.ABS_H
                            sD0 = 0 if D0 < block.BLOCK_ABS_D[k] else D0 - block.BLOCK_ABS_D[k]
                            sD1 = block.BLOCK_SIZE[k] if d1 > <int32_t>(block.BLOCK_ABS_D[k] + block.BLOCK_SIZE[k]) \
                                else d1 - block.BLOCK_ABS_D[k]

                            # vertices of buffer block
                            bV0 = 0 if V0 > intersect_area.V0 else intersect_area.V0 - V0
                            bH0 = 0 if H0 > intersect_area.H0 else intersect_area.H0 - H0
                            bD0 = 0 if D0 > block.BLOCK_ABS_D[k] else block.BLOCK_ABS_D[k] - D0

                            if b"NULL.tif" in slice_fullpath:
                                continue

                            fmt_mngr.copy_file_block2buffer(slice_fullpath,
                                                            sV0, sV1, sH0, sH1, sD0, sD1,
                                                            subvol, sbv_bytes_chan,
                                                            bH0 + bV0 * sbv_width + bD0 * sbv_width * sbv_height,
                                                            sbv_width,
                                                            sbv_width * sbv_height,
                                                            sbv_width * sbv_height * sbv_depth)
        else:
            raise IOError("TiledVolume: Depth interval out of range")

        return npsubvol



cimport numpy as np
import numpy as np
from libc.math cimport sqrt, exp, pow
import cython

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef neuron_meanshift(tree: list[tuple], img: np.ndarray, window_radius=32, order=2.):
    """
    profile swc radius based on an image.

    :param tree: list of swc nodes 
    :param img: 3D neuronal image
    :param window_radius: the max radius of meanshift window
    :param order: the order to scale the intensity
    :return: 
    """
    assert img.ndim == 3
    tree_ = [list(t) for t in tree]
    cdef float x, y, z, r
    cdef np.ndarray[np.float32_t, ndim=3] img_ = img.astype(np.float32)
    for t in tree_:
        x, y, z = t[2:5]
        marker_meanshift(x, y, z, img_, window_radius, order)
        t[2:5] = x, y, z
    return [tuple(t) for t in tree]


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef profile_radius(tree: list[tuple], img: np.ndarray, is2d: bool, bkg_thr: float):
    """
    profile swc radius based on an image.
    
    :param tree: list of swc nodes 
    :param img: 3D neuronal image
    :param is2d: 2D radius profiling
    :param bkg_thr: background threshold
    :return: 
    """
    assert img.ndim == 3
    tree_ = [list(t) for t in tree]
    cdef float x, y, z, r
    cdef np.ndarray[np.float32_t, ndim=3] img_
    if img.dtype == np.float32:
        img_ = img
    else:
        img_ = img.astype(np.float32)
    for t in tree_:
        x, y, z = t[2:5]
        if is2d:
            t[5] = marker_radius_hanchuan_xy(x, y, z, img_, bkg_thr)
        else:
            t[5] = marker_radius_hanchuan(x, y, z, img_, bkg_thr)
    return [tuple(t) for t in tree]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double marker_radius_hanchuan(float x, float y, float z, np.ndarray[np.float32_t, ndim=3] img, double thr):
    cdef long long sz0 = img.shape[2], sz1 = img.shape[1], sz2 = img.shape[0], i, j, k
    cdef double max_r = sz0 / 2, total_num, background_num, ir, dz, dy, dx, zlower, zupper, r
    max_r = min(max_r, sz1 / 2)
    max_r = min(max_r, sz2 / 2)

    for ir in range(1, <int>(max_r + 1)):
        total_num = background_num = 0

        zlower = -ir
        zupper = ir
        for dz in range(<int>zlower, <int>(zupper + 1)):
            for dy in range(<int>-ir, <int>(ir + 1)):
                for dx in range(<int>-ir, <int>(ir + 1)):
                    total_num += 1
                    r = sqrt(dx*dx + dy*dy + dz*dz)
                    if ir - 1 < r <= ir:
                        i = <long long>(x + dx)
                        if i < 0 or i >= sz0:
                            return ir
                        j = <long long>(y + dy)
                        if j < 0 or j >= sz1:
                            return ir
                        k = <long long>(z + dz)
                        if k < 0 or k >= sz2:
                            return ir
                        if img[k, j, i] <= thr:
                            background_num += 1
                            if background_num / total_num > 0.001:
                                return ir
    return ir


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double marker_radius_hanchuan_xy(float x, float y, float z, np.ndarray[np.float32_t, ndim=3] img, double thr):
    cdef long long sz0 = img.shape[2], sz1 = img.shape[1], i, j, k = <long long>z
    cdef double max_r = sz0 / 2, total_num, background_num, ir, dy, dx, r
    max_r = min(max_r, sz1 / 2)

    for ir in range(1, <int>(max_r + 1)):
        total_num = background_num = 0

        for dy in range(<int>-ir, <int>(ir + 1)):
            for dx in range(<int>-ir, <int>(ir + 1)):
                total_num += 1
                r = sqrt(dx*dx + dy*dy)
                if ir - 1 < r <= ir:
                    i = <long long>(x + dx)
                    if i < 0 or i >= sz0:
                        return ir
                    j = <long long>(y + dy)
                    if j < 0 or j >= sz1:
                        return ir
                    if img[k, j, i] <= thr:
                        background_num += 1
                        if background_num / total_num > 0.001:
                            return ir
    return ir


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void marker_meanshift(float & x, float & y, float & z, np.ndarray[np.float32_t, ndim=3] img, int window_radius, double order):
    cdef:
        int wr_t = window_radius
        long long X, Y, Z, dX, dY, dZ
        double mean = 0., count = 0., w, center_dist, s2, prefactor, tot_x, tot_y, tot_z, sum_v, s, x_, y_, z_, dist
    for X in range( <int>max(x+.5-wr_t, 0), <int>min(img.shape[2] - 1, x + .5 + wr_t) ):
        for Y in range(<int> max(x + .5 - wr_t, 0), <int> min(img.shape[1] - 1, x + .5 + wr_t)):
            for Z in range(<int> max(x + .5 - wr_t, 0), <int> min(img.shape[0] - 1, x + .5 + wr_t)):
                dX = <int>(X - x)
                dY = <int>(Y - y)
                dZ = <int>(Z - z)
                dist = sqrt(dX*dX + dY*dY + dZ*dZ)
                if dist > wr_t:
                    continue
                w = pow(img[Z, Y, X], order)
                mean += w * dist
                count += w
    mean /= count + 1e-5
    wr_t = <int>max(1., mean)
    s = wr_t / 4.
    s2 = pow(s, 2)
    prefactor = 1. / pow(2 * np.pi * s2, 1.5)
    while True:
        tot_x = tot_y = tot_z = sum_v = 0.
        for X in range(<int> max(x + .5 - wr_t, 0), <int> min(img.shape[2] - 1, x + .5 + wr_t)):
            for Y in range(<int> max(x + .5 - wr_t, 0), <int> min(img.shape[1] - 1, x + .5 + wr_t)):
                for Z in range(<int> max(x + .5 - wr_t, 0), <int> min(img.shape[0] - 1, x + .5 + wr_t)):
                    dX = <int>(X - x)
                    dY = <int>(Y - y)
                    dZ = <int>(Z - z)
                    dist = sqrt(dX * dX + dY * dY + dZ * dZ)
                    if dist > wr_t:
                        continue
                    w = pow(img[Z, Y, X], order) * prefactor * exp(-dist / (2 * s2))
                    tot_x += w * X
                    tot_y += w * Y
                    tot_z += w * Z
                    sum_v += w
        if sum_v < 1e-5:
            break
        x_ = tot_x / sum_v
        y_ = tot_y / sum_v
        z_ = tot_z / sum_v
        center_dist = sqrt( (x-x_)*(x-x_) + (y-y_)*(y-y_) + (z-z_)*(z-z_) )
        x = x_
        y = y_
        z = z_
        wr_t -= <int>s
        if center_dist < 1:
            break


# cpdef neuron_radius_terafly(tree: list[tuple], img_path: str, is2d: bool, block_size: int, meanshift: bool, bkg_thr: float)